from torch import Tensor

from quaterion.loss.similarity_loss import SimilarityLoss


class PairwiseLoss(SimilarityLoss):
    """Base class for pairwise losses.

    Args:
        distance_metric_name: Name of the function, that returns a distance between two
            embeddings. The class SiameseDistanceMetric contains pre-defined metrics
            that can be used.
    """

    def __init__(self, distance_metric_name: str = "cosine_distance"):
        super(PairwiseLoss, self).__init__(distance_metric_name=distance_metric_name)

    def forward(
        self,
        embeddings: Tensor,
        pairs: Tensor,
        labels: Tensor,
        subgroups: Tensor,
    ) -> Tensor:
        """Compute loss value.

        Args:
            embeddings: shape: (batch_size, vector_length)
            pairs: shape: (2 * pairs_count,) - contains a list of known similarity pairs
                in batch
            labels: shape: (pairs_count,) - similarity of the pair
            subgroups: shape: (2 * pairs_count,) - subgroup ids of objects

        Returns:
            Tensor: zero-size tensor, loss value
        """
        raise NotImplementedError()



from typing import Collection, Hashable, Any, List, Callable, Union, Tuple, Iterable

from torch import Tensor

from quaterion_models.encoders import Encoder
from quaterion_models.types import TensorInterchange


KeyExtractorType = Callable[[Any], Hashable]

CacheCollateReturnType = Union[List[Hashable], Tuple[Hashable, TensorInterchange]]

CacheCollateFnType = Callable[
    [Collection[Any]],
    CacheCollateReturnType,
]


class CacheEncoder(Encoder):
    """Wrapper for encoders to avoid repeated calculations.

    Encoder results can be calculated one time and reused after in situations when
    encoder's layers are frozen and provide deterministic embeddings for input data.

    Args:
        encoder: Encoder object to be wrapped.
        key_extractor: function required to extract hashable key from complicated
            objects which can't be hashed with default key extractor.
    """

    def __init__(self, encoder: Encoder, key_extractor: KeyExtractorType = None):
        if encoder.trainable():
            raise ValueError("Trainable encoder can't be cached")
        super().__init__()
        self._encoder = encoder
        self.cache_filled = False
        self.key_extractor = (
            key_extractor if key_extractor is not None else self.default_key_extractor
        )

    def trainable(self) -> bool:
        """Defines if encoder is trainable. This flag affects caching and checkpoint
        saving of the encoder.

        Returns:
            bool: whether encoder trainable or not
        """
        return False

    def embedding_size(self) -> int:
        """Size of output embedding.

        Returns:
            int: Size of resulting embedding.
        """
        return self._encoder.embedding_size()

    @classmethod
    def default_key_extractor(cls, obj: Any) -> Hashable:
        """Default implementation of key extractor.

        Generate hashable from batch object with built-in hash, also support dicts.

        Returns:
             Hashable: Key for cache
        """
        return (
            hash(obj) if not isinstance(obj, dict) else hash(tuple(sorted(obj.items())))
        )

    def key_collate_fn(self, batch: Collection[Any]) -> List[Hashable]:
        """Convert batch of raw data into list of keys for cache.

        Returns:
             List[Hashable]: List of keys for cache
        """
        return [self.key_extractor(value) for value in batch]

    def cache_collate(
        self, batch: Collection[Any]
    ) -> Union[List[Hashable], Tuple[Hashable, TensorInterchange]]:
        """Converts raw data batch into suitable model input and keys for caching.

        Returns:
             Union[
                List[Hashable],
                Tuple[Hashable, TensorInterchange]
            ]: Tuple of keys and according model input during caching process, only
            list of keys after cache has been filled
        """
        keys: List[Hashable] = self.key_collate_fn(batch)
        if self.cache_filled:
            return keys
        values: TensorInterchange = self._encoder.get_collate_fn()(batch)
        return keys, values

    def get_collate_fn(self) -> CacheCollateFnType:
        """Provides function that converts raw data batch into suitable model input.

        Returns:
             CacheCollateFnType: cache collate function
        """
        return self.cache_collate

    def forward(self, batch: TensorInterchange) -> Tensor:
        """Infer encoder.

        Convert input batch to embeddings

        Args:
            batch: collated batch (currently, it can be only batch of keys)
        Returns:
            Tensor: shape: (batch_size, embedding_size) - embeddings
        """
        raise NotImplementedError()

    def save(self, output_path: str):
        """Persist current state to the provided directory

        Args:
            output_path: path to save to
        """
        self._encoder.save(output_path)

    @classmethod
    def load(cls, input_path: str) -> Encoder:
        """CachedEncoder classes wrap already instantiated encoders and don't
        provide loading support.

        Args:
            input_path: path to load from
        """
        raise ValueError("Cached encoder does not support loading")

    def fill_cache(self, data: Tuple[Iterable[Hashable], TensorInterchange]) -> None:
        """Apply wrapped encoder to data and store processed data on
        corresponding device.

        Args:
            data: Tuple of keys and batches suitable for encoder

        """
        raise NotImplementedError()

    def reset_cache(self):
        """Reset all stored data."""
        raise NotImplementedError()



from typing import Union

import torch


def info_value_of_dtype(dtype: torch.dtype) -> Union[torch.finfo, torch.iinfo]:
    """Returns the `finfo` or `iinfo` object of a given PyTorch data type.

    Does not allow torch.bool.

    Args:
        dtype: dtype for which to return info value

    Returns:
        Union[torch.finfo, torch.iinfo]: info about given data type

    Raises:
        TypeError: if torch.bool is passed
    """
    if dtype == torch.bool:
        raise TypeError("Does not support torch.bool")
    elif dtype.is_floating_point:
        return torch.finfo(dtype)
    else:
        return torch.iinfo(dtype)


def min_value_of_dtype(dtype: torch.dtype) -> Union[int, float]:
    """Returns the minimum value of a given PyTorch data type.

    Does not allow torch.bool.

    Args:
        dtype: dtype for which to return min value

    Returns:
        Union[int, float]: max value of dtype
    """
    return info_value_of_dtype(dtype).min


def max_value_of_dtype(dtype: torch.dtype) -> float:
    """Returns the maximum value of a given PyTorch data type.

    Does not allow torch.bool.

    Args:
        dtype: dtype for which to return max value

    Returns:
        Union[int, float]: max value of dtype
    """
    return info_value_of_dtype(dtype).max



from torch import Tensor, LongTensor

from quaterion.loss.similarity_loss import SimilarityLoss


class GroupLoss(SimilarityLoss):
    """Base class for Group losses.

    Args:
        distance_metric_name: Name of the function, that returns a distance between
            two embeddings. The class SiameseDistanceMetric contains pre-defined metrics
            that can be used.
    """

    def __init__(self, distance_metric_name: str = "cosine_distance"):
        super(GroupLoss, self).__init__(distance_metric_name=distance_metric_name)

    def forward(self, embeddings: Tensor, groups: LongTensor) -> Tensor:
        """

        Args:
            embeddings: shape: (batch_size, vector_length)
            groups: shape: (batch_size,) - Groups, associated with embeddings

        Returns:
            Tensor: zero-size tensor, loss value
        """
        raise NotImplementedError()



from enum import Enum
from dataclasses import dataclass, field
from typing import Dict, Optional, Callable, Any, Hashable


class CacheType(str, Enum):
    """Available tensor devices to be used for caching.

    AUTO: use CUDA if it is available, else use CPU.
    CPU: tensors device is CPU.
    GPU: tensors device is GPU.
    """

    AUTO = "auto"
    CPU = "cpu"
    GPU = "gpu"


@dataclass
class CacheConfig:
    """Class to be used in `configure_cache` of `TrainableModel`.

    cache_type: cache type for single encoder which has no custom name set. In all other
        cases use `mapping`.
    mapping: mapping of `encoder_name` to `CacheType`, only specified inhere encoders
        will be cached, the others will stay untouched.
    key_extractors: mapping of encoders to key extractor functions required to cache
        non-hashable objects.
    batch_size: batch size to be used in CacheDataLoader during caching process. It does
        not affect others training stages.
    num_workers: num_workers to be used in CacheDataLoader during caching process. It
        does not affect others training stages.
    """

    cache_type: Optional[CacheType] = None
    mapping: Dict[str, CacheType] = field(default_factory=dict)
    key_extractors: Dict[str, Callable[[Any], Hashable]] = field(default_factory=dict)
    batch_size: Optional[int] = 32
    num_workers: Optional[int] = None  # if None - inherited from source dl



from typing import Type, Dict, Any

from torch import Tensor, relu, LongTensor

from quaterion.loss.metrics import SiameseDistanceMetric
from quaterion.loss.pairwise_loss import PairwiseLoss
from quaterion.utils import max_value_of_dtype


class ContrastiveLoss(PairwiseLoss):
    """Contrastive loss.

    Expects as input two texts and a label of either 0 or 1. If the label == 1, then the
    distance between the two embeddings is reduced. If the label == 0, then the distance
    between the embeddings is increased.

    Further information:
        http://yann.lecun.com/exdb/publis/pdf/hadsell-chopra-lecun-06.pdf

    Args:
        distance_metric_name: Name of the function, that returns a distance between two
            embeddings. The class SiameseDistanceMetric contains pre-defined metrics
            that can be used.
        margin: Negative samples (label == 0) should have a distance of at least the
            margin value.
        size_average: Average by the size of the mini-batch.
    """

    @classmethod
    def metric_class(cls) -> Type:
        """Class with metrics available for current loss.

        Returns:
            Type: class containing metrics
        """
        return SiameseDistanceMetric

    def __init__(
        self,
        distance_metric_name: str = "cosine_distance",
        margin: float = 1.0,
        size_average: bool = True,
    ):
        super().__init__(distance_metric_name=distance_metric_name)
        self.margin = margin
        self.size_average = size_average

    def get_config_dict(self) -> Dict[str, Any]:
        """Config used in saving and loading purposes.

        Config object has to be JSON-serializable.

        Returns:
            Dict[str, Any]: JSON-serializable dict of params
        """
        return {
            **super().get_config_dict(),
            "margin": self.margin,
            "size_average": self.size_average,
        }

    def forward(
        self,
        embeddings: Tensor,
        pairs: LongTensor,
        labels: Tensor,
        subgroups: Tensor,
        **kwargs
    ):
        """Compute loss value.

        Args:
            embeddings: Batch of embeddings, first half of embeddings are embeddings
                of first objects in pairs, second half are embeddings of second objects
                in pairs.
            pairs: Indices of corresponding objects in pairs.
            labels: Scores of positive and negative objects.
            subgroups: subgroups to distinguish objects which can and cannot be used
                as negative examples
            **kwargs: additional key-word arguments for generalization of loss call

        Returns:
            Tensor: averaged or summed loss value
        """
        rep_anchor = embeddings[pairs[:, 0]]
        rep_other = embeddings[pairs[:, 1]]
        distances = self.distance_metric(rep_anchor, rep_other)
        negative_distances_impact = 0.0

        if len(subgroups.unique()) > 1:
            # shape (2 * batch_size, embeddings_size)
            embeddings_count = embeddings.shape[0]  # `embeddings_count` consists of
            # number of embeddings for `obj_a` and `obj_b`

            # `subgroups` shape is (embeddings_count,)
            # shape (embeddings_count, embeddings_count)
            subgroup_matrix: Tensor = subgroups.repeat(embeddings_count, 1)
            # shape (embeddings_count, embeddings_count)
            comp_matrix: Tensor = subgroup_matrix != subgroup_matrix.T
            # a matrix to take into account only distances to negative
            # examples, i.e. from examples which don't belong to current
            # subgroup

            # shape (embeddings_count, embeddings_count)
            distance_matrix = self.distance_metric(embeddings, embeddings, matrix=True)
            distance_matrix[~comp_matrix] = max_value_of_dtype(distance_matrix.dtype)
            # shape (embeddings_count, 1)
            negative_distances, _ = distance_matrix.min(dim=1)  # find negative examples
            # which are the closest to positive ones
            # shape (embeddings_count // 2, 1)
            neg_dist_to_anchors = negative_distances[pairs[:, 0]]
            # shape (embeddings_count // 2, 1)
            neg_dist_to_other = negative_distances[pairs[:, 1]]
            # shape (embeddings_count // 2, 1)
            negative_distances_impact = relu(self.margin - neg_dist_to_anchors).pow(
                2
            ) + relu(self.margin - neg_dist_to_other).pow(2)

        # shape (embeddings_count // 2, 1)
        losses = (
            0.5
            * (
                labels.float() * distances.pow(2)
                + (1 - labels).float() * relu(self.margin - distances).pow(2)
            )
            + negative_distances_impact
        )

        return losses.mean() if self.size_average else losses.sum()



from typing import Type, Callable, Dict, Any

from torch import nn, Tensor

from quaterion.loss.metrics import SiameseDistanceMetric


class SimilarityLoss(nn.Module):
    """Base similarity losses class.

    Args:
        distance_metric_name: Name of the function, that returns a distance between two
            embeddings. The class SiameseDistanceMetric contains pre-defined metrics
            that can be used.
    """

    def __init__(self, distance_metric_name: str = "cosine_distance"):
        super(SimilarityLoss, self).__init__()
        self.distance_metric_name = distance_metric_name
        self.distance_metric = self.get_distance_function(self.distance_metric_name)

    @classmethod
    def metric_class(cls) -> Type:
        """Class with metrics available for current loss.

        Returns:
            Type: class containing metrics
        """
        return SiameseDistanceMetric

    @classmethod
    def get_distance_function(
        cls, function_name: str
    ) -> Callable[[Tensor, Tensor], Tensor]:
        """Retrieve distance function from metric class.

        Args:
            function_name: name of attribute in cls.metric_class instance

        Returns:
            Callable[[Tensor, Tensor], Tensor]: function for distance calculation

        Raises:
            RuntimeError: unknown `distance_metric_name` is passed
        """
        for name, foo in vars(cls.metric_class()).items():
            if name == function_name:
                return foo.__func__

        raise RuntimeError(
            f"Unknown `distance_metric_name` {function_name},"
            f" available metrics: {vars(cls.metric_class()).keys()}"
        )

    def get_config_dict(self) -> Dict[str, Any]:
        """Config used in saving and loading purposes.

        Config object has to be JSON-serializable.

        Returns:
            Dict[str, Any]: JSON-serializable dict of params
        """
        return {"distance_metric_name": self.distance_metric_name}



from typing import Any, List, Generic, Dict, Tuple

import torch
from torch.utils.data import DataLoader
from torch.utils.data.dataloader import T_co

from quaterion.dataset.similarity_samples import (
    SimilarityPairSample,
    SimilarityGroupSample,
)


class SimilarityDataLoader(DataLoader, Generic[T_co]):
    @classmethod
    def fetch_unique_objects(cls, batch: List[Any]) -> List[Any]:
        """Fetch unique objects to avoid calculation of repeated objects embeddings.

        Args:
            batch: batch of raw data

        Returns:
            List[Any]: list of unique objects

        """
        raise NotImplementedError()


class PairsSimilarityDataLoader(SimilarityDataLoader[SimilarityPairSample]):
    @classmethod
    def collate_fn(
        cls, batch: List[SimilarityPairSample]
    ) -> Tuple[List[Any], Dict[str, torch.Tensor]]:
        """Collate function for SimilarityPairSamples objects.

        Extract features from pairs and collects them in list.
        Construct labels dict with `pairs`, `labels` and `subgroups`.

        Args:
            batch: List of SimilarityPairSample objects

        Returns:
            Tuple[List[Any], Dict[str, torch.Tensor]]: tuple of features and labels

        Examples:
            ```
            PairsSimilarityDataLoader.collate_fn(
                [
                    SimilarityPairSample(
                        obj_a="1st_pair_1st_obj", obj_b="1st_pair_2nd_obj", score=1.0,
                    ),
                    SimilarityPairSample(
                        obj_a="2nd_pair_1st_obj",
                        obj_b="2nd_pair_2nd_obj",
                        score=0.0,
                        subgroup=1
                    ),
                ]
            )

            # result
            (
                # features
                [
                    '1st_pair_1st_obj',
                    '2nd_pair_1st_obj',
                    '1st_pair_2nd_obj',
                    '2nd_pair_2nd_obj'
                ],
                # labels
                {
                    'labels': tensor([1., 0.]),
                    'pairs': tensor([[0, 2], [1, 3]]),
                    'subgroups': tensor([0., 1., 0., 1.])
                }
            )
            ```
        """
        features = [record.obj_a for record in batch] + [
            record.obj_b for record in batch
        ]
        labels = {
            "pairs": torch.LongTensor([[i, i + len(batch)] for i in range(len(batch))]),
            "labels": torch.Tensor([record.score for record in batch]),
            "subgroups": torch.Tensor([record.subgroup for record in batch] * 2),
        }
        return features, labels

    @classmethod
    def fetch_unique_objects(cls, batch: List[SimilarityPairSample]) -> List[Any]:
        """Fetch unique objects from SimilarityPairSample batch.

        Collect unique `obj_a` and `obj_b` from samples in a batch.

        Args:
            batch: List of SimilarityPairSample's

        Returns:
            List[Any]: list of unique `obj_a` and `obj_b` in batch
        """
        unique_objects = []
        for sample in batch:
            if sample.obj_a not in unique_objects:
                unique_objects.append(sample.obj_a)
            if sample.obj_b not in unique_objects:
                unique_objects.append(sample.obj_b)
        return unique_objects


class GroupSimilarityDataLoader(SimilarityDataLoader[SimilarityGroupSample]):
    @classmethod
    def collate_fn(
        cls, batch: List[SimilarityGroupSample]
    ) -> Tuple[List[Any], Dict[str, torch.Tensor]]:
        """Collate function for SimilarityGroupSamples objects.

        Extract features from pairs and collects them in list.
        Construct labels dict with `groups`.

        Args:
            batch: List of SimilarityGroupSample objects

        Returns:
            Tuple[List[Any], Dict[str, torch.Tensor]]: tuple of features and
                labels

        Examples:
            ```
            GroupSimilarityDataLoader.collate_fn(
                [
                    SimilarityGroupSample(
                        obj="orange",
                        group=0,
                    ),
                    SimilarityGroupSample(
                        obj="lemon",
                        group=0,
                    ),
                    SimilarityGroupSample(
                        obj="apple",
                        group=1,
                    )
                ]
            )

            # result
            (
                # features
                ['orange', 'lemon', 'apple'],
                # labels
                {'groups': tensor([0, 0, 1])}
            )
            ```
        """
        features = [record.obj for record in batch]
        labels = {"groups": torch.LongTensor([record.group for record in batch])}
        return features, labels

    @classmethod
    def fetch_unique_objects(cls, batch: List[SimilarityGroupSample]) -> List[Any]:
        """Fetch unique objects from SimilarityGroupSample batch.

        Collect unique `obj` from samples in a batch.

        Args:
            batch: List of SimilarityGroupSample's

        Returns:
            List[Any]: list of unique `obj` in batch
        """
        unique_objects = []
        for sample in batch:
            if sample.obj not in unique_objects:
                unique_objects.append(sample.obj)
        return unique_objects



import torch

from torch import Tensor, pairwise_distance, cosine_similarity
from torch.nn import functional as F


class SiameseDistanceMetric:
    """The metric for the contrastive loss."""

    @staticmethod
    def euclidean(x: Tensor, y: Tensor, matrix=False) -> Tensor:
        """Compute euclidean distance

        Args:
            x: shape: (batch_size, ...)
            y: shape: (batch_size, ...)
            matrix: flat to calculate distance matrix (all to all)
        Returns:
            Tensor: shape (batch_size, 1)
        """
        if not matrix:
            return pairwise_distance(x, y, p=2)
        raise NotImplementedError()

    @staticmethod
    def manhattan(x: Tensor, y: Tensor, matrix=False) -> Tensor:
        """Compute manhattan distance

        Args:
            x: shape: (batch_size, ...)
            y: shape: (batch_size, ...)
            matrix: flat to calculate distance matrix (all to all)
        Returns:
            Tensor: shape (batch_size, 1)
        """
        if not matrix:
            return pairwise_distance(x, y, p=1)
        raise NotImplementedError()

    @staticmethod
    def cosine_distance(x: Tensor, y: Tensor, matrix=False) -> Tensor:
        """Compute cosine distance

        Args:
            x: shape: (batch_size, ...)
            y: shape: (batch_size, ...)
            matrix: flat to calculate distance matrix (all to all)
        Returns:
            Tensor: shape (batch_size, 1)
        """

        if not matrix:
            return 1 - cosine_similarity(x, y)

        x_norm = F.normalize(x, p=2, dim=1)
        y_norm = F.normalize(y, p=2, dim=1).transpose(0, 1)
        return 1 - torch.mm(x_norm, y_norm)

    @staticmethod
    def dot_product_distance(x: Tensor, y: Tensor, matrix=False) -> Tensor:
        """Compute dot product distance

        Dot product distance may have unexpected effects. Make sure you entirely
        understand how it works itself, and with chosen loss function especially.

        Args:
            x: shape: (batch_size, ...)
            y: shape: (batch_size, ...)
            matrix: flat to calculate distance matrix (all to all)
        Returns:
            Tensor: shape (batch_size, 1)
        """
        return (
            -torch.einsum("id,jd->ij", x, y)
            if matrix
            else -torch.einsum("id,id->i", x, y)
        )



from typing import Dict, Any, Union, Optional

import pytorch_lightning as pl

from torch import Tensor
from pytorch_lightning.utilities.types import (
    TRAIN_DATALOADERS,
    EVAL_DATALOADERS,
)

from quaterion_models.encoders import Encoder
from quaterion_models.heads import EncoderHead
from quaterion_models import MetricModel
from quaterion_models.types import TensorInterchange
from quaterion.train.encoders import (
    CacheConfig,
    CacheType,
)
from quaterion.loss import SimilarityLoss
from quaterion.utils.enums import TrainStage
from quaterion.train.cache_mixin import CacheMixin


class TrainableModel(pl.LightningModule, CacheMixin):
    """Base class for models to be trained."""

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

        encoders = self.configure_encoders()
        self.cache_config = self.configure_caches()
        encoders = self._apply_cache_config(encoders, self.cache_config)

        head = self.configure_head(MetricModel.get_encoders_output_size(encoders))

        self._model = MetricModel(encoders=encoders, head=head)
        self._loss = self.configure_loss()

    @property
    def model(self) -> MetricModel:
        """Origin model to be trained

        Returns:
            MetricModel: model to be trained
        """
        return self._model

    @property
    def loss(self) -> SimilarityLoss:
        """Property to get the loss function to use."""
        return self._loss

    def configure_loss(self) -> SimilarityLoss:
        """Method to configure loss function to use."""
        raise NotImplementedError()

    def configure_encoders(self) -> Union[Encoder, Dict[str, Encoder]]:
        """Method to provide encoders configuration

        Use this function to define an initial state of encoders.
        This function should be used to assign initial values for encoders
        before training as well as during the checkpoint loading.

        Returns:
             Union[Encoder, Dict[str, Encoder]]: one instance of encoder which will
                have quaterion_models.model.DEFAULT_ENCODER_KEY, or dict of encoder
                names and corresponding encoder instances.
        """
        raise NotImplementedError()

    def configure_caches(self) -> Optional[CacheConfig]:
        """Method to provide cache configuration

                Use this method to define which encoders should cache calculated embeddings and
                what kind of cache they should use.

                Returns:
                    Optional[CacheConfig]: cache configuration to be applied if provided, None
                        otherwise
                Examples:

                >>> CacheConfig(CacheType.AUTO)
                CacheConfig(
                    cache_type=<CacheType.AUTO: 'auto'>,
                    mapping={},
                    key_extractors={}
                )

                >>> cache_config = CacheConfig(
        ...     mapping={"text_encoder": CacheType.GPU, "image_encoder": CacheType.CPU}
        ... )
                CacheConfig(
                    cache_type=None,
                    mapping={
                        'text_encoder': <CacheType.GPU: 'gpu'>,
                        'image_encoder': <CacheType.CPU: 'cpu'>
                    },
                    key_extractors={}
                )
                >>> CacheConfig(
        ...     cache_type=CacheType.AUTO,
        ...     key_extractors={"default": lambda obj: hash(obj)}
        ... )
                CacheConfig(
                    cache_type=<CacheType.AUTO: 'auto'>,
                    mapping={},
                    key_extractors={'default': <function <lambda> at 0x106bc90e0>}
                )

        """
        pass

    def configure_head(self, input_embedding_size: int) -> EncoderHead:
        """Use this function to define an initial state for head layer of the model.

        Args:
            input_embedding_size: size of embeddings produced by encoders
        Returns:
            EncoderHead: instance of EncoderHead to be added to the model
        """
        raise NotImplementedError()

    def process_results(
        self,
        embeddings: Tensor,
        targets: Dict[str, Any],
        batch_idx: int,
        stage: TrainStage,
        **kwargs,
    ):
        """Method to provide any additional evaluations of embeddings.

        Args:
            embeddings: shape: (batch_size, embedding_size) - model's output.
            targets: output of batch target collate.
            batch_idx: ID of the processing batch.
            stage: train, validation or test stage.
        """
        pass

    def training_step(
        self, batch: TensorInterchange, batch_idx: int, **kwargs
    ) -> Tensor:
        """Compute and return the training loss and some additional metrics for e.g.
        the progress bar or logger.

        Args:
            batch: Output of DataLoader.
            batch_idx: Integer displaying index of this batch.
            **kwargs: keyword arguments to be passed into `process_results`

        Returns:
            Tensor: computed loss value
        """
        stage = TrainStage.TRAIN
        loss = self._common_step(
            batch=batch, batch_idx=batch_idx, stage=stage, **kwargs
        )
        return loss

    def validation_step(self, batch, batch_idx, **kwargs) -> Optional[Tensor]:
        """Compute validation loss and some additional metrics for e.g. the progress
        bar or logger.

        Args:
            batch: Output of DataLoader.
            batch_idx: Integer displaying index of this batch.
            **kwargs: keyword arguments to be passed into `process_results`
        """
        stage = TrainStage.VALIDATION
        self._common_step(batch=batch, batch_idx=batch_idx, stage=stage, **kwargs)
        return None

    def test_step(self, batch, batch_idx, **kwargs) -> Optional[Tensor]:
        """Compute test loss and some additional metrics for e.g. the progress
        bar or logger.

        Args:
            batch: Output of DataLoader.
            batch_idx: Integer displaying index of this batch.
            **kwargs: keyword arguments to be passed into `process_results`
        """
        stage = TrainStage.TEST
        self._common_step(batch=batch, batch_idx=batch_idx, stage=stage, **kwargs)
        return None

    def _common_step(self, batch, batch_idx, stage: TrainStage, **kwargs) -> Tensor:
        """Common step to compute loss and metrics for training, validation, test
         and other stages.

        Args:
            batch: Output of DataLoader.
            batch_idx: Integer displaying index of this batch.
            stage: current training stage: training, validation, etc.
            **kwargs: keyword arguments to be passed into `process_results`

        Returns:
            Tensor: computed loss value
        """
        features, targets = batch
        embeddings = self.model(features)
        loss = self.loss(embeddings=embeddings, **targets)
        self.log(f"{stage}_loss", loss)
        self.process_results(
            embeddings=embeddings,
            targets=targets,
            batch_idx=batch_idx,
            stage=stage,
            **kwargs,
        )
        return loss

    def save_servable(self, path: str):
        """Save model for serving, independent of Pytorch Lightning

        Args:
            path: path to save to
        """
        self.model.save(path)

    # region anchors
    # https://github.com/PyTorchLightning/pytorch-lightning/issues/10667
    def train_dataloader(self) -> TRAIN_DATALOADERS:
        pass

    def test_dataloader(self) -> EVAL_DATALOADERS:
        pass

    def val_dataloader(self) -> EVAL_DATALOADERS:
        pass

    def predict_dataloader(self) -> EVAL_DATALOADERS:
        pass

    # endregion
